;====================================================================
; NVIDIA NForce2 custom option ROM for ASRock K7NF2-RAID
; vR.C
; author: digitalbath
; 20 Aug 2021
; Credits to tictac, pinczakko, polygon, TzK, Antinomy and Infrared
; Version History
; vR.0: new Version
; vR.C: new build with new items
;====================================================================
use16                                           ; 16bit mode
ROM_SIZE_IN_BLOCK = 6                           ; 1 means ROM size is 1 block (512 bytes)
ROM_SIZE_IN_BYTE = ROM_SIZE_IN_BLOCK * 512      ; number of 512 byte blocks

define PCI_ADDR_PORT    0xCF8
define PCI_DATA_PORT    0xCFC
define CMOS_ADDR_PORT   0x70
define CMOS_DATA_PORT   0x71

macro configurePort pci
{
    mov eax, pci                                ; move address for 32bit register offset
    mov dx, PCI_ADDR_PORT                       ; pci register address port
    out dx, eax                                 ; set port we want to read
}

macro readPciReg pci
{
    configurePort pci                           ; pci register address port
    mov dx, PCI_DATA_PORT                       ; pci register data port
    in eax, dx                                  ; read register values
}


macro writePciReg pci, data, mask
{
    readPciReg pci                              ; pci register data port
    and eax, mask                               ; mask for data to remove pre-set value
    or eax, data                                ; add custom value to old register data
    out dx, eax                                 ; write new custom register value
}

macro readCmosReg reg, mask
{
    mov al, reg                                 ; set CMOS index
    out CMOS_ADDR_PORT, al                      ; send register offset
    in al, CMOS_DATA_PORT                       ; read CMOS data
    and al, mask                                ; mask the data and zero unneeded bits
}

ROMStart:
    db 0x055, 0x0AA                             ; ROM Header 55,AA -> Bootable rom
    db (ROMEnd - ROMStart)/512                  ; ROM Size in 512 bytes
    jmp MAIN                                    ; jump to main routine
    times 	0x18-($-$$) db 0	        ; zero fill in between
    dw	PCI_DATA_STRUC	                        ; Pointer to PCI HDR structure (at 18h)
    times 	0x1A-($-$$) db 0	        ; zero fill in between
    dw	PnP_Header	                        ; PnP Expansion Header Pointer (at 1Ah)


;====================================================================
; PCI data structure
;
PCI_DATA_STRUC:
	db	'PCIR'		                ; PCI Header Sign
	dw	0x10DE		                ; Vendor ID
	dw	0x008F		                ; Device ID
	dw	0x00		                ; VPD
	dw	0x18		                ; PCI data struc length (byte)
	db	0x00	                	; PCI Data struct Rev
	db	0x02		                ; Base class code, 02h == Network Controller
	db	0x00		                ; Sub class code = 00h and interface = 00h -->Ethernet Controller
	db	0x00		                ; Interface code, see PCI Rev2.2 Spec Appendix D
	dw	ROM_SIZE_IN_BLOCK               ; Image length in mul of 512 byte, little endian format
	dw	0x00		                ; rev level
	db	0x00		                ; Code type = x86
	db	0x80		                ; last image indicator
	dw	0x00		                ; reserved

;====================================================================
; PnP ROM Bios Header
;
PnP_Header:
	db	'$PnP'		         	; PnP Rom header sign
	db	0x01		        	; Structure Revision
	db	0x02		        	; Header structure Length in mul of 16 bytes
	dw	0x00		         	; Offset to next header (00 if none)
	db	0x00		        	; reserved
	db	0x7A		         	; 8 Bit checksum (for this header, -->
				        	;  --> check again after compile and repair if needed)
	dd	0x00		         	; PnP Device ID --> 0h in Realtek RPL ROM
	dw	Manufacturer_str	        ; pointer to manufacturer string
	dw	Product_str		        ; pointer to product string
	db	0x02,0x00,0x00		        ; Device Type code 3 byte
	db	0x14		        	; Device Indicator, 14h from RPL ROM-->See Page 18 of
				        	; PnP BIOS spec., Lo nibble (4) means IPL device

	dw	0x00			        ; Boot Connection Vector, 00h = disabled
	dw	0x00		         	; Disconnect Vector, 00h = disabled
	dw	MAIN    	         	; Bootstrap Entry Vector (BEV)
	dw	0x00		        	; reserved
	dw	0x00		         	; Static resource Information vector (0000h if unused)

;====================================================================
; Identifier strings
;
Manufacturer_str  db	'digitalbath',00h

Product_str	db	'PCI option ROM',00h

;====================================================================
;PCI Option ROM initialization Code (init function)
;
MAIN:
    ; save all registers for later restore
    pushfd
    pusha
;
;====================================================================
;
; CMOS register and mask used in 1B system module:
;
;---------------------------------------------------------------------------
; Cmos register 40h 0x1Fh  T(RC) Bus: 0 Device: 0 Function: 1 Mem Controller
; Register 90h 8-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 0000.0001b =  9T
;   2: 0000.0010b = 10T
;   3: 0000.0011b = 11T
;   4: 0000.0100b = 12T
;   5: 0000.0101b = 13T
;   6: 0000.0110b = 14T
;
;----------------------------------------------------------------------------
; Cmos register 41h 0x1Fh T(RFC)  Bus: 0 Device: 0 Function: 1 Mem Controller
; Register 91h 8-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 0000.0001b = 12T
;   2: 0000.0010b = 13T
;   3: 0000.0011b = 14T
;   4: 0000.0100b = 15T
;   5: 0000.0101b = 16T
;   6: 0000.0110b = 17T
;
;---------------------------------------------------------------------------
; Cmos register 48h 0x0Fh   Drive Strength Bus: 0 Device: 0 Function: 4 Mem Controller
;
; Dimm 1 Drive Strength, Registers 7D, 81   8-bit
; Dimm 2 Drive Strength, Registers 67, 73   8-bit
; Dimm 3 Drive Strength, Registers 65, 71   8-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 0000.0001b = 1T (11h)
;   2: 0000.0010b = 2T (22h)
;   3: 0000.0011b = 3T (33h)
;   4: 0000.0100b = 4T (44h)
;   5: 0000.0101b = 5T (55h)
;   6: 0000.0110b = 6T (66h)
;
;---------------------------------------------------------------------------
; Cmos register 4Bh 0x0Fh  Slew Rate Bus: 0 Device: 0 Function: 4 Mem Controller
; Dimm 1 Slew Rate, Register 7C 8-bit
; Dimm 2 Slew Rate, Register 66 8-bit
; Dimm 3 Slew Rate, Register 64 8-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 0000.0001b =  6T (66h)
;   2: 0000.0010b =  7T (77h)
;   3: 0000.0011b =  8T (88h)
;   4: 0000.0100b =  9T (99h)
;   5: 0000.0101b = 10T (AAh)
;   6: 0000.0110b = 11T (BBh)
;
;---------------------------------------------------------------------------
; Cmos register 42h 0x0Fh    alpha timings  Bus: 0 Device: 0 Function: 1 Mem Controller
;  Register 97h F0h 8-bit T(DOE)
;  Register 97h 0Fh 8-bit T(RRD)
;  Register 96h F0h 8-bit T(W2P)
;  Register 96h 0Fh 8-bit T(W2R)
;  Register 95h F0h 8-bit T(REXT)
;  Register 95h 0Fh 8-bit T(R2P)
;  Register 94h F0h 8-bit T(TR2W)
;
;   0: Auto (keep value, skip custom code)
;   1: 0000.0001b = 1-3-3-3-3-3-4: 97h:13 96h:33 95h:33 94h:40
;   2: 0000.0010b = 1-3-4-3-3-4-4: 97h:13 96h:43 95h:34 94h:40
;   3: 0000.0011b = 2-3-3-3-3-3-4: 97h:23 96h:33 95h:33 94h:40
;   4: 0000.0100b = 2-3-5-3-2-2-3: 97h:23 96h:53 95h:22 94h:30
;   5: 0000.0101b = 3-2-5-3-2-2-3: 97h:32 96h:53 95h:22 94h:30
;   6: 0000.0110b = 3-4-4-4-3-4-5: 97h:34 96h:44 95h:34 94h:50
;   7: 0000.0111b = 3-4-5-4-3-4-5: 97h:34 96h:54 95h:34 94h:50
;   8: 0000.1000b = 4-4-5-4-3-3-5: 97h:44 96h:54 95h:33 94h:50
;   9: 0000.1001b = 4-5-5-5-3-5-6: 97h:45 96h:55 95h:35 94h:60
;   A: 0000.1010b = 5-4-6-3-3-5-6: 97h:54 96h:63 95h:35 94h:60
;   B: 0000.1011b = 5-6-4-3-3-5-6: 97h:56 96h:43 95h:35 94h:60
;
;--------------------------------------------------------------------------
; Cmos register 46h 0x1Fh  T(REF)   Bus: 0 Device: 0 Function: 1 Mem Controller
; Register 60h+61h 16-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 0000.0001b = 1806 -> 0618h = 1560 cycles
;   2: 0000.0010b = 1807 -> 0718h = 1816 cycles
;   3: 0000.0011b = 0008 -> 0800h = 2048 cycles
;   4: 0000.0100b = 1008 -> 0810h = 2064 cycles
;   5: 0000.0101b = 2009 -> 0920h = 2336 cycles
;   6: 0000.0110b = 000A -> 0A00h = 2560 cycles
;   7: 0000.0111b = 200A -> 0A20h = 2592 cycles
;   8: 0000.1000b = 000C -> 0C00h = 3072 cycles
;   9: 0000.1001b = 300C -> 0C30h = 3120 cycles
;   A: 0000.1010b = 300E -> 0E30h = 3632 cycles
;   B: 0000.1011b = 640E -> 0E64h = 3684 cycles
;
;
;====================================================================
;
    readCmosReg 0x40, 0x1F                      ; read tRC from CMOS write to al

    cmp al, 0
    je trc_end                                  ; jump to next setting if auto selected (equal 0000.0000)

    xor ebx, ebx
;
;
;trc_9
    cmp al, 0x01
    je .trc_9
;
;trc_10
    cmp al, 0x02
    je .trc_10
;
;trc_11
    cmp al, 0x03
    je .trc_11
;
;trc_12
    cmp al, 0x04
    je .trc_12
;
;trc_13
    cmp al, 0x05
    je .trc_13
;
;trc_14
    cmp al, 0x06
    je .trc_14
;
;
    .trc_9:
    mov  bl, 0x09                               ; Value 9 (09h)
    je .end3
;
    .trc_10:
    mov bl, 0x0A                                ; Value 10 (10h)
    je .end3
;
    .trc_11:
    mov bl, 0x0B                                ; Value 11 (0Bh)
    je .end3
;
    .trc_12:
    mov bl, 0x0C                                ; Value 12 (0Ch)
    je .end3
;
    .trc_13:
    mov bl, 0x0D                                ; Value 13 (0Dh)
    je .end3
;
    .trc_14:
    mov bl, 0x0E                                ; Value 14 (0Eh)
    je .end3
;
;
  .end3:
    writePciReg 0x80000190, ebx, 0xFFFFFF00     ; offset 90
    je trc_end
;
    trc_end:					; jump target

;====================================================================

    readCmosReg 0x41, 0x1F                      ; read tRFC from CMOS

    cmp al, 0
    je trfc_end                                 ; jump to next setting if auto selected (equal 0000.0000)

    xor ebx, ebx
;
;
;trfc_12
    cmp al, 0x01
    je .trfc_12
;
;trfc_13
    cmp al, 0x02
    je .trfc_13
;
;trfc_14
    cmp al, 0x03
    je .trfc_14
;
;trfc_15
    cmp al, 0x04
    je .trfc_15
;
;trfc_16
    cmp al, 0x05
    je .trfc_16
;
;trfc_17
    cmp al, 0x06
    je .trfc_17
;
;
    .trfc_12:
    mov  bh, 0x0C                               ; Value 12 (0Ch)
    je .end4
;
    .trfc_13:
    mov  bh, 0x0D                               ; Value 13 (0Dh)
    je .end4
;
    .trfc_14:
    mov  bh, 0x0E                               ; Value 14 (0Eh)
    je .end4
;
    .trfc_15:
    mov  bh, 0x0F                               ; Value 15 (0Fh)
    je .end4
;
    .trfc_16:
    mov  bh, 0x10                               ; Value 16 (10h)
    je .end4
;
    .trfc_17:
    mov  bh, 0x11                               ; Value 17 (11h)
    je .end4
;
;
  .end4:
    writePciReg 0x80000190, ebx, 0xFFFFE0FF     ;offset 91
    je trfc_end
;
    trfc_end:                                   ; jump target
;
;====================================================================
    readCmosReg 0x4B, 0x0F                      ; read slew rate from CMOS

    cmp al, 0
    je slew_end                                 ; jump to next setting if auto selected (equal 0000.0000)

    xor ebx, ebx
;
;
;slew_6
    cmp al, 0x01
    je .slew_6
;
;slew_7
    cmp al, 0x02
    je .slew_7
;
;slew_8
    cmp al, 0x03
    je .slew_8
;
;slew_9
    cmp al, 0x04
    je .slew_9
;
;slew_A
    cmp al, 0x05
    je .slew_A
;
;slew_B
    cmp al, 0x06
    je .slew_B
;
;
    .slew_6:
    writePciReg 0x80000464, 0x00660066, 0xFF00FF00     ; offset 64 and 66, value 66h
    writePciReg 0x8000047C, 0x00000066, 0xFFFFFF00     ; offset 7C, 66h
    jmp slew_end
;
    .slew_7:
    writePciReg 0x80000464, 0x00770077, 0xFF00FF00     ; offset 64 and 66, value 77h
    writePciReg 0x8000047C, 0x00000077, 0xFFFFFF00     ; offset 7C, 77h
    jmp slew_end
;
    .slew_8:
    writePciReg 0x80000464, 0x00880088, 0xFF00FF00     ; offset 64 and 66, value 88h
    writePciReg 0x8000047C, 0x00000088, 0xFFFFFF00     ; offset 7C, 88h
    jmp slew_end
;
    .slew_9:
    writePciReg 0x80000464, 0x00990099, 0xFF00FF00     ; offset 64 and 66, value 99h
    writePciReg 0x8000047C, 0x00000099, 0xFFFFFF00     ; offset 7C, 99h
    jmp slew_end
;
    .slew_A:
    writePciReg 0x80000464, 0x00AA00AA, 0xFF00FF00     ; offset 64 and 66, value AAh
    writePciReg 0x8000047C, 0x000000AA, 0xFFFFFF00     ; offset 7C, AAh
    jmp slew_end 
;
    .slew_B:
    writePciReg 0x80000464, 0x00BB00BB, 0xFF00FF00     ; offset 64 and 66, value BBh
    writePciReg 0x8000047C, 0x000000BB, 0xFFFFFF00     ; offset 7C, BBh
    jmp slew_end
; 
;
    slew_end:                                   ; jump target

;====================================================================

    readCmosReg 0x48, 0x0F                      ; read drive strength from CMOS

    cmp al, 0
    je drive_end                                ; jump to next setting if auto selected (equal 0000.0000)

    xor ebx, ebx
;
;
;drive_1
    cmp al, 0x01
    je .drive_1
;
;drive_2
    cmp al, 0x02
    je .drive_2
;
;drive_3
    cmp al, 0x03
    je .drive_3
;
;drive_4
    cmp al, 0x04
    je .drive_4
;
;drive_5
    cmp al, 0x05
    je .drive_5
;
;drive_6
    cmp al, 0x06
    je .drive_6
;
;
    .drive_1:
    writePciReg 0x80000464, 0x11001100, 0x00FF00FF     ; offset 65 and 67, value 11h
    writePciReg 0x80000470, 0x11001100, 0x00FF00FF     ; offset 71 and 73, value 11,
    writePciReg 0x8000047C, 0x00001100, 0xFFFF00FF     ; offset 7D, value 11h
    writePciReg 0x80000480, 0x00001100, 0xFFFF00FF     ; offset 81, value 11h
    jmp drive_end
;
    .drive_2:
    writePciReg 0x80000464, 0x22002200, 0x00FF00FF     ; offset 65 and 67, value 22h
    writePciReg 0x80000470, 0x22002200, 0x00FF00FF     ; offset 71 and 73, value 22,
    writePciReg 0x8000047C, 0x00002200, 0xFFFF00FF     ; offset 7D, value 22h
    writePciReg 0x80000480, 0x00002200, 0xFFFF00FF     ; offset 81, value 22h
    jmp drive_end
;
    .drive_3:
    writePciReg 0x80000464, 0x33003300, 0x00FF00FF     ; offset 65 and 67, value 33h
    writePciReg 0x80000470, 0x33003300, 0x00FF00FF     ; offset 71 and 73, value 33,
    writePciReg 0x8000047C, 0x00003300, 0xFFFF00FF     ; offset 7D, value 33h
    writePciReg 0x80000480, 0x00003300, 0xFFFF00FF     ; offset 81, value 33h
    jmp drive_end
;
    .drive_4:
    writePciReg 0x80000464, 0x44004400, 0x00FF00FF     ; offset 65 and 67, value 44h
    writePciReg 0x80000470, 0x44004400, 0x00FF00FF     ; offset 71 and 73, value 44,
    writePciReg 0x8000047C, 0x00004400, 0xFFFF00FF     ; offset 7D, value 44h
    writePciReg 0x80000480, 0x00004400, 0xFFFF00FF     ; offset 81, value 44h
    jmp drive_end
;
    .drive_5:
    writePciReg 0x80000464, 0x55005500, 0x00FF00FF     ; offset 65 and 67, value 55h
    writePciReg 0x80000470, 0x55005500, 0x00FF00FF     ; offset 71 and 73, value 55,
    writePciReg 0x8000047C, 0x00005500, 0xFFFF00FF     ; offset 7D, value 55h
    writePciReg 0x80000480, 0x00005500, 0xFFFF00FF     ; offset 81, value 55h
    jmp drive_end
;
    .drive_6:
    writePciReg 0x80000464, 0x66006600, 0x00FF00FF     ; offset 65 and 67, value 66h
    writePciReg 0x80000470, 0x66006600, 0x00FF00FF     ; offset 71 and 73, value 66,
    writePciReg 0x8000047C, 0x00006600, 0xFFFF00FF     ; offset 7D, value 66h
    writePciReg 0x80000480, 0x00006600, 0xFFFF00FF     ; offset 81, value 66h
    jmp drive_end
;
;
    drive_end:                                  ; jump target
;====================================================================


    

    readCmosReg 0x42, 0x0F                      ; read alpha from CMOS

    cmp al, 0
    je alpha_end                                ; jump to end if auto selected (equal 0000.0000)

    xor ebx, ebx

    cmp al, 0x01
    je .alpha1                                  ;jump if equal 0000.0001
;
    cmp al, 0x02
    je .alpha2                                  ;jump if equal 0000.0010
;
    cmp al, 0x03
    je .alpha3                                  ;jump if equal 0000.0011
;
    cmp al, 0x04
    je .alpha4                                  ;jump if equal 0000.0100
;
    cmp al, 0x05
    je .alpha5                                  ;jump if equal 0000.0101
;
    cmp al, 0x06
    je .alpha6                                  ;jump if equal 0000.0110
;
    cmp al, 0x07
    je .alpha7                                  ;jump if equal 0000.0111
;
    cmp al, 0x08
    je .alpha8                                  ;jump if equal 0000.1000
;
    cmp al, 0x09
    je .alpha9                                  ;jump if equal 0000.1001
;
    cmp al, 0x0A
    je .alpha10                                 ;jump if equal 0000.1010
;
    cmp al, 0x0B
    je .alpha11                                 ;jump if equal 0000.1011
;
;
  .alpha1:
    mov ebx, 0x13333340                         ;97h:13 96h:33 95h:33 94h:40
    je .end1
  .alpha2:
    mov ebx, 0x13433440                         ;97h:13 96h:43 95h:34 94h:40
    je .end1
  .alpha3:
    mov ebx, 0x23333340                         ;97h:23 96h:33 95h:33 94h:40
    je .end1
  .alpha4:
    mov ebx, 0x23532230                         ;97h:23 96h:53 95h:22 94h:30
    je .end1
  .alpha5:
    mov ebx, 0x32532230                         ;97h:32 96h:53 95h:22 94h:30
    je .end1
  .alpha6:
    mov ebx, 0x34443450                         ;97h:34 96h:44 95h:34 94h:50
    je .end1
  .alpha7:
    mov ebx, 0x34543450                         ;97h:34 96h:54 95h:34 94h:50
    je .end1
  .alpha8:
    mov ebx, 0x44543350                         ;97h:44 96h:54 95h:33 94h:50
    je .end1
  .alpha9:
    mov ebx, 0x45553560                         ;97h:45 96h:55 95h:35 94h:60
    je .end1
  .alpha10:
    mov ebx, 0x54633560                         ;97h:54 96h:63 95h:35 94h:60
    je .end1
  .alpha11:
    mov ebx, 0x56433560                         ;97h:56 96h:43 95h:35 94h:60 
    je .end1
;
;
  .end1:
    writePciReg 0x80000194, ebx, 0x0000000F
    je alpha_end

    alpha_end:                                  ; jump target
;
;====================================================================
;
    readCmosReg 0x46, 0x1F                      ; read tREF from CMOS

    cmp al, 0
    je tref_end                                 ; jump to end if auto selected (equal 0000.0000)

    xor ebx, ebx

    cmp al, 0x01
    je .tref1                                   ;jump if equal 0000.0001
;
    cmp al, 0x02
    je .tref2                                   ;jump if equal 0000.0010
;
    cmp al, 0x03
    je .tref3                                   ;jump if equal 0000.0011
;
    cmp al, 0x04
    je .tref4                                   ;jump if equal 0000.0100
;
    cmp al, 0x05
    je .tref5                                   ;jump if equal 0000.0101
;
    cmp al, 0x06
    je .tref6                                   ;jump if equal 0000.0110
;
    cmp al, 0x07
    je .tref7                                   ;jump if equal 0000.0111
;
    cmp al, 0x08
    je .tref8                                   ;jump if equal 0000.1000
;
    cmp al, 0x09
    je .tref9                                   ;jump if equal 0000.1001
;
    cmp al, 0x0A
    je .tref10                                  ;jump if equal 0000.1010
;
    cmp al, 0x0B
    je .tref11                                  ;jump if equal 0000.1011
;
;
  .tref1:
    mov ebx, 0x0618                             ; 1560 Cycles
    je .end2
  .tref2:
    mov ebx, 0x0718                             ; 1816 Cycles
    je .end2
  .tref3:
    mov ebx, 0x0800                             ; 2048 Cycles
    je .end2
  .tref4:
    mov ebx, 0x0810                             ; 2064 Cycles
    je .end2
  .tref5:
    mov ebx, 0x0920                             ; 2336 Cycles
    je .end2
  .tref6:
    mov ebx, 0x0A00                             ; 2560 Cycles
    je .end2
  .tref7:
    mov ebx, 0x0A20                             ; 2592 Cycles
    je .end2
  .tref8:
    mov ebx, 0x0C00                             ; 3072 Cycles
    je .end2
  .tref9:
    mov ebx, 0x0C30                             ; 3120 Cycles
    je .end2
  .tref10:
    mov ebx, 0x0E30                             ; 3632 Cycles
    je .end2
  .tref11:
    mov ebx, 0x0E64                             ; 3684 Cycles
    je .end2

  .end2:
    writePciReg 0x80000160, ebx, 0xFFFF0000
    je tref_end

    tref_end:                                  ; jump target

;====================================================================
;   
    jmp codeend
;
;====================================================================

codeend:
    popa
    popfd
    retf                                        ; return far to system bios routine

                                        
  times (ROM_SIZE_IN_BYTE-$) db 0               ; use 00h as the padding bytes until we reach the ROM size
                                                ; The last byte (512th) will be the patch_byte for the checksum
                                                ; patch_byte is calculated and automagically inserted below
  PREV_CHKSUM = 0
  repeat $
  load CHKSUM byte from %-1
  CHKSUM = (PREV_CHKSUM + CHKSUM) mod 0x100
  PREV_CHKSUM = CHKSUM
  end repeat
  
  store byte (0x100 - CHKSUM) at ($-1)          ; store the patch_byte
ROMEnd:
